import osmnx as ox
import networkx as nx
import folium
import itertools
import pickle
import time, datetime
import os.path
import sys
import webbrowser

import random
 
import os
import glob


from shapely.geometry import LineString
from shapely.geometry import MultiLineString
from shapely.geometry import MultiPolygon
from shapely.geometry import Point
from shapely.geometry import Polygon
from shapely.ops import linemerge

import geopandas as gpd


import matplotlib.pyplot as plt
plt.switch_backend('agg')


ox.settings.log_console=True
ox.settings.use_cache=True

print('OSMnx version: ', ox.__version__)
print('NetworkX version: ',nx.__version__)
print('Python version: ',sys.version)

start_time = time.time()

water_tags = {'natural':['spring','water'],
              'water':['pond','lake','reservoir'],
              'landuse':['reservoir','basin'],
              'amenity':['watering_place','drinking_water','water'],
              'waterway':'stream',
              'man_made':'water_tank'
        }



def elapsed_time(start):
    end_time = time.time()
    duration = round(end_time - start)
    timestamp = str(datetime.timedelta(seconds = duration))
    print(timestamp)

def graph_length(Q):
    length = ox.stats.edge_length_total(Q) #in meters
    length = length * 0.0006213712
    return length

def get_bounds(Q):
    QD = Q.to_directed()
    QD.graph['crs'] = ox.settings.default_crs    
    nodes, edges = ox.graph_to_gdfs(QD)
    tb = edges.total_bounds
    print( tb )
    sw = [tb[1],tb[0]]
    ne = [tb[3],tb[2]]
    return [sw, ne]

def report(Q, txt=''):
    print(len(Q.nodes()), ' nodes ', txt)
    print(len(Q.edges()), ' edges ', txt)
    if isinstance(Q,nx.MultiGraph):    
        length = ox.stats.edge_length_total(Q) #in meters
        length = length * 0.0006213712
        print(f'total length {length:.2f} miles ', txt)
    zeroes = 0    
    odds = 0
    evens = 0
    twos = 0
    ones = 0
    fours = 0
    four_pairs = 0
    parallels = 0
    for node in Q.nodes:
        d = Q.degree(node)
        if d == 0:
            zeroes += 1
        if d % 2 == 1:
            odds += 1
        else:
            evens += 1
        if d == 2:
            twos += 1
        if d == 1:
            ones += 1
        if d==4:
            fours += 1
            for neighbor in Q.neighbors(node):
                if Q.degree(neighbor)==4 and neighbor>node: # prevent double-count
                    four_pairs += 1
        if d>6:
            print('node ', node, 'has degree ', d)
        for neighbor in Q.neighbors(node):
            edge_count = Q.number_of_edges(node,neighbor)
            if edge_count > 1:
                parallels += edge_count - 1
                
    if zeroes > 0:            
        print('number of zero-degree nodes ' + txt + ' :', zeroes)            
    print('number of odd intersections ' + txt + ' :', odds)
    print('number of even intersections ' + txt + ' :', evens)
    print('number of 2-way intersections ' + txt + ' :', twos)
    print('number of 4-way intersections ' + txt + ' :', fours)
    print('number of trailhead/terminus ' + txt + ' :', ones)
    print('number of parallel edges ' + txt + ' :', parallels)
    if four_pairs > 0:
        print('number of 4-way adjacent pairs ' + txt + ' :', four_pairs)

        
    #check_node_geometry(Q)
    print('')

def colorize_graph(Q):
    QR = Q.copy()
    for node, dat in Q.nodes(data=True):
        d = Q.degree(node)
        dat['color'] = d

        QR.add_node(node,**dat)                #replace node value

    for u, v, k, dat in Q.edges(keys=True,data=True):
        d = 'path'
        if 'highway' in dat:
            d = dat['highway']
        dat['color']=0.0
        if isinstance(d, str):
            if d=='path' or d == 'footway':
                dat['color']=1.0
            else:
                dat['color'] = 2.0
        QR.add_edge(u,v,key=k,**dat)
    QR.graph['color']=True
    return QR
    

town_markers = [(34.29963, -108.13268, 'Pie Town'),
         (33.19593, -108.20834, 'Gila Hot Springs'),
         (32.77981, -108.27533, 'Silver City'),
         (33.37939, -108.90369, 'Alma/Glenwood'),
         (33.03022, -108.16864, 'Lake Roberts'),
         (32.85594, -107.98024, 'Mimbres'),       
         (32.91685, -107.70593, 'Kingston')]

# draw() changed after moving to Linux
def draw(Q,title='',show_nodes = False, boxes=None, not_boxes = None, markers=town_markers):
    start_latlng = (33.04876, -108.30680)
    m = folium.Map(location=[start_latlng[1],start_latlng[0]],
               zoom_start=12)
    # convert to MultiDiGraph for folium
    #Q1,Q2 = split_merged(Q)
    T = Q.copy()
    if not 'color' in T.graph:
        T = colorize_graph(T)
    QD = T.to_directed()
 #   QD.graph['crs'] = ox.settings.default_crs

##    for u,v,k,d in QD.edges(data=True, keys=True):
##        for attr in d:
##            if type(d[attr]) is list:
##                print(d)
        

    # our graphs may be HUGE, so simplify if not already done so
    if not "simplified" in QD.graph:
        QD = ox.simplify_graph(QD, edge_attrs_differ=['name'], endpoint_attrs=None, remove_rings=False, track_merged=True)
#        QD = ox.simplify_graph(QD, edge_attrs_differ=None, endpoint_attrs=None, remove_rings=True, track_merged=True)
#        QD = simplify_graph_modified(QD)
    QD = colorize_graph(QD)

    
    nodes, edges = ox.graph_to_gdfs(QD)
    edges.info()
    #print(edges)
 #   edges.to_html(open('edge_gdf.html', 'w'))
 #   webbrowser.open('edge_gdf.html')
#    m = edges.explore(color='#ff00ff', tiles="openstreetmap",tooltip='name')
    m = edges.explore(column='color',  cmap = 'tab10',legend=False, tiles="openstreetmap",tooltip='name')
#    m = edges.explore(column='color',  cmap = 'tab10',legend=False, tiles="openstreetmap")
    if show_nodes:    
#        nodes.explore(m=m, color="pink", marker_kwds={"radius": 4})    
        nodes.explore(m=m, column='color', cmap='tab10', legend=False,marker_kwds={"radius": 4})    

    if not boxes == None:
        for n,s,e,w in boxes:
            folium.Rectangle(bounds=[(n,e),(n,w),(s,e),(s,w)], color='#ff7800', fill=True, fill_color='#ffff00', fill_opacity=0.2).add_to(m)
    if not not_boxes == None:
        for n,s,e,w in not_boxes:
            folium.Rectangle(bounds=[(n,e),(n,w),(s,e),(s,w)], color='#00ff78', fill=False, fill_color='#00ffff', fill_opacity=0.2).add_to(m)
    if not markers == None:
        kw = {"prefix": "fa", "color": "green", "icon": "arrow-up"}
        icon = folium.Icon(angle=0, **kw)
        for lat,lon, txt in markers:
            #print(lat,lon,txt)
            folium.Marker(location=[lat,lon],popup=txt,
                          icon=folium.Icon(color='lightred',icon=''),
                          tooltip=txt).add_to(m)
            
    # include the option to switch map layers
   # folium.TileLayer('openstreetmap').add_to(m)
    folium.TileLayer('cartodbpositron').add_to(m)
   # folium.TileLayer('Stamen Terrain').add_to(m)
    folium.TileLayer(
        tiles='https://{s}.tile-cyclosm.openstreetmap.fr/cyclosm/{z}/{x}/{y}.png',
        name='cyclosm',
        attr='CyclOSM',
    ).add_to(m)
    folium.LayerControl().add_to(m)    

    title_html = '''
            <h3 align="center" style="font-size:16px"><b>{}</b></h3>
             '''.format(title)   
    m.get_root().html.add_child(folium.Element(title_html))
    file_name = "/home/sagebrush/biz/graph/map_" + title + ".html"
    m.save(file_name)
    #m.show_in_browser()
    webbrowser.open(file_name)
    input("Press Enter to continue...")
    print('*************', flush=True)
    print("", flush=True)    


def draw_gdf(gdf,title='',show_nodes=True,boxes=None, not_boxes = None, markers=town_markers):
    start_latlng = (33.04876, -108.30680)
    m = folium.Map(location=[start_latlng[1],start_latlng[0]],
               zoom_start=12)
    gdf_nodes = gdf.copy()

    gdf_nodes = gpd.GeoDataFrame(
        gdf, geometry=gpd.points_from_xy(gdf.x, gdf.y), crs=ox.settings.default_crs  
        )
    if show_nodes:    
        m = gdf_nodes.explore(color='blue', legend=False, tiles="openstreetmap",tooltip=True,marker_kwds={"radius": 4})    

    if not boxes == None:
        for n,s,e,w in boxes:
            folium.Rectangle(bounds=[(n,e),(n,w),(s,e),(s,w)], color='#ff7800', fill=True, fill_color='#ffff00', fill_opacity=0.2).add_to(m)
    if not not_boxes == None:
        for n,s,e,w in not_boxes:
            folium.Rectangle(bounds=[(n,e),(n,w),(s,e),(s,w)], color='#00ff78', fill=False, fill_color='#00ffff', fill_opacity=0.2).add_to(m)
    if not markers == None:
        kw = {"prefix": "fa", "color": "green", "icon": "arrow-up"}
        icon = folium.Icon(angle=0, **kw)
        for lat,lon, txt in markers:
            #print(lat,lon,txt)
            folium.Marker(location=[lat,lon],popup=txt,
                          icon=folium.Icon(color='lightred',icon=''),
                          tooltip=txt).add_to(m)
            
    # include the option to switch map layers
   # folium.TileLayer('openstreetmap').add_to(m)
    folium.TileLayer('cartodbpositron').add_to(m)
   # folium.TileLayer('Stamen Terrain').add_to(m)
    folium.TileLayer(
        tiles='https://{s}.tile-cyclosm.openstreetmap.fr/cyclosm/{z}/{x}/{y}.png',
        name='cyclosm',
        attr='CyclOSM',
    ).add_to(m)
    folium.LayerControl().add_to(m)    

    title_html = '''
            <h3 align="center" style="font-size:16px"><b>{}</b></h3>
             '''.format(title)   
    m.get_root().html.add_child(folium.Element(title_html))
    file_name = "/home/sagebrush/biz/graph/map_" + title + ".html"
    m.save(file_name)
    #m.show_in_browser()
    webbrowser.open(file_name)
    input("Press Enter to continue...")
    print('*************', flush=True)
    print("", flush=True)



### ###########  FIND WATER SOURCES    #############
bbox_gila = (34.35704, 32.77935, -107.65983, -108.94574) # rough box of Gila NF
wgdf = ox.features.features_from_bbox(bbox=bbox_gila, tags=water_tags)



# I really did not want to learn about GeoPandas, but there you go...
water_points = wgdf.copy()
print('crs = ', ox.settings.default_crs)
#water_points = water_points.to_crs(crs=ox.settings.default_crs)
water_points = ox.projection.project_gdf(water_points,to_latlong=True)
water_points['geometry'] = water_points['geometry'].centroid

water_points['x'] = None
water_points['y'] = None
water_points['x'] = water_points.geometry.apply(lambda z: z.x)
water_points['y'] = water_points.geometry.apply(lambda z: z.y)

water_points.info()
print(water_points.head())

graph_attrs = {'crs' : ox.settings.default_crs}
WG = nx.MultiGraph(**graph_attrs)

for index, row in water_points.iterrows():
    data = row.to_dict()
    #print(data)
    l = []
    for key in data.keys():
        if key=='nodes' or key=='x' or key=='y' or key=='geometry' or key=='source':
            continue
        if not isinstance(data[key],str):
            continue                          
        if data[key]=='nan':
            continue
        l.append(key + '=' + data[key])
    s = ','.join(l)
    data['name'] = s
    #https://stackoverflow.com/questions/31433989/return-copy-of-dictionary-excluding-specified-keys
    new_data = {x: data[x] for x in data if x in ['x','y','name']}
    WG.add_node(index[1],**new_data) 
 #   print(index[1],data)
    

        
report(WG,'water sources')


water_points = ox.io._stringify_nonnumeric_cols(water_points)
water_points.info()
print(water_points.head())

#draw_gdf(water_points,'gila water sources', show_nodes = True)

# This is too many water sources. Let us filter by proximity to our route.

with open('max_candidate_with_towns.pkl', 'rb') as f:
    J = pickle.load(f)
JD = J.to_directed()  #nearest_edges expects directed
report(J,'simplified, loaded from disk')
if nx.is_eulerian(J):
    print('saved graph is eulerian!!!!!')
else:
    print('back to the drawing board')

# help from https://stackoverflow.com/questions/69844440/incorrect-results-using-osmnx-nearest-edges

WJ = WG.copy()

E,D = ox.distance.nearest_edges(JD,X=water_points['x'],Y=water_points['y'],return_dist=True)

max_distance = 2000 #meters
EARTH_RADIUS_M = 6_371_009 # new convenience notation for big numbers 
max_distance = max_distance / EARTH_RADIUS_M

# Not clear from documentation, but nearest_edges returns distance in units of
# earth radius if projection is lat/long.
for node,d in zip(WG.nodes(),D):
    if d > max_distance:
        WJ.remove_node(node)

print('J node data')
for i,nd in zip(range(10),J.nodes(data=True)):
    data=nd[1]
    print(nd[0],data)

print('WJ node data')
for i,nd in zip(range(10),WJ.nodes(data=True)):
    data=nd[1]
    print(nd[0],data)
    
# Brace yourself for One Ugly Hack.
for u,v,dat in J.edges(data=True): #add one dummy edge to WJ for draw() to work
    WJ.add_node(u, ** J.nodes[u])
    WJ.add_node(v, ** J.nodes[v])
    WJ.add_edge(u,v,**dat)
    break
   

report(WJ,'water near route')        

print('number of water nodes: ', WJ.number_of_nodes())
print('number of J nodes ', J.number_of_nodes())
for node, data in WJ.nodes(data=True):
    J.add_node(node,**data)
print('number of J nodes after adding', J.number_of_nodes())
for i,nd in zip(range(10),WJ.nodes(data=True)):
    data=nd[1]
    print(data)

draw(WJ,'water sources near route', show_nodes=True)
    
draw(J,'route with water sources', show_nodes=True)
#draw_gdf(water_points,'water')


    

elapsed_time(start_time)

print('finished script ' + os.path.basename(__file__))

